/*
 * gulkan
 * Copyright 2018 Collabora Ltd.
 * Author: Lubosz Sarnecki <lubosz.sarnecki@collabora.com>
 * SPDX-License-Identifier: MIT
 */

#include "gulkan-instance.h"

#include <vulkan/vulkan.h>

gboolean
gulkan_has_error (VkResult     res,
                  const gchar *fun,
                  const gchar *file,
                  int          line)
{
  if (res != VK_SUCCESS)
  {
    g_printerr ("ERROR: %s failed with %s in %s:%d\n",
                fun, vk_result_string (res), file, line);
    return TRUE;
  }
  return FALSE;
}

#define ENUM_TO_STR(r) case r: return #r

const gchar*
vk_result_string (VkResult code)
{
  switch (code)
    {
      ENUM_TO_STR(VK_SUCCESS);
      ENUM_TO_STR(VK_NOT_READY);
      ENUM_TO_STR(VK_TIMEOUT);
      ENUM_TO_STR(VK_EVENT_SET);
      ENUM_TO_STR(VK_EVENT_RESET);
      ENUM_TO_STR(VK_INCOMPLETE);
      ENUM_TO_STR(VK_ERROR_OUT_OF_HOST_MEMORY);
      ENUM_TO_STR(VK_ERROR_OUT_OF_DEVICE_MEMORY);
      ENUM_TO_STR(VK_ERROR_INITIALIZATION_FAILED);
      ENUM_TO_STR(VK_ERROR_DEVICE_LOST);
      ENUM_TO_STR(VK_ERROR_MEMORY_MAP_FAILED);
      ENUM_TO_STR(VK_ERROR_LAYER_NOT_PRESENT);
      ENUM_TO_STR(VK_ERROR_EXTENSION_NOT_PRESENT);
      ENUM_TO_STR(VK_ERROR_FEATURE_NOT_PRESENT);
      ENUM_TO_STR(VK_ERROR_INCOMPATIBLE_DRIVER);
      ENUM_TO_STR(VK_ERROR_TOO_MANY_OBJECTS);
      ENUM_TO_STR(VK_ERROR_FORMAT_NOT_SUPPORTED);
      ENUM_TO_STR(VK_ERROR_SURFACE_LOST_KHR);
      ENUM_TO_STR(VK_ERROR_NATIVE_WINDOW_IN_USE_KHR);
      ENUM_TO_STR(VK_SUBOPTIMAL_KHR);
      ENUM_TO_STR(VK_ERROR_OUT_OF_DATE_KHR);
      ENUM_TO_STR(VK_ERROR_INCOMPATIBLE_DISPLAY_KHR);
      ENUM_TO_STR(VK_ERROR_VALIDATION_FAILED_EXT);
      ENUM_TO_STR(VK_ERROR_INVALID_SHADER_NV);
      default:
        return "UNKNOWN RESULT";
    }
}

/**
 * GulkanInstance:
 *
 * A #GulkanInstance-struct which represents the Vulkan instance.
 */
struct _GulkanInstance
{
  GObjectClass parent_class;

  VkInstance instance;
};

G_DEFINE_TYPE (GulkanInstance, gulkan_instance, G_TYPE_OBJECT)

static void
gulkan_instance_init (GulkanInstance *self)
{
  self->instance = VK_NULL_HANDLE;
}

GulkanInstance *
gulkan_instance_new (void)
{
  return (GulkanInstance*) g_object_new (GULKAN_TYPE_INSTANCE, 0);
}

static void
gulkan_instance_finalize (GObject *gobject)
{
  GulkanInstance *self = GULKAN_INSTANCE (gobject);
  vkDestroyInstance (self->instance, NULL);
}

static void
gulkan_instance_class_init (GulkanInstanceClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);
  object_class->finalize = gulkan_instance_finalize;
}

static gboolean
_init_instance_extensions (GSList   *required_extensions,
                           char    **enabled_extensions,
                           uint32_t *out_num_enabled)
{
  uint32_t num_enabled = 0;
  uint32_t num_extensions = 0;
  VkResult res =
    vkEnumerateInstanceExtensionProperties (NULL, &num_extensions, NULL);
  vk_check_error ("vkEnumerateInstanceExtensionProperties", res, FALSE);

  VkExtensionProperties *extension_props =
    g_malloc (sizeof(VkExtensionProperties) * num_extensions);

  if (num_extensions > 0)
  {
    res = vkEnumerateInstanceExtensionProperties (NULL, &num_extensions,
                                                  extension_props);
    vk_check_error ("vkEnumerateInstanceExtensionProperties", res, FALSE);

    for (size_t i = 0; i < g_slist_length (required_extensions); i++)
    {
      gboolean found = FALSE;
      for (uint32_t j = 0; j < num_extensions; j++)
      {
        GSList* extension_name = g_slist_nth (required_extensions, (guint) i);
        if (strcmp ((gchar*) extension_name->data,
                    extension_props[j].extensionName) == 0)
        {
          found = TRUE;
          enabled_extensions[num_enabled++] =
            g_strdup (extension_props[j].extensionName);
          break;
        }
      }

      if (!found)
      {
        GSList* extension_name = g_slist_nth (required_extensions, (guint) i);
        g_printerr ("Vulkan missing requested extension '%s'.\n",
                    (gchar*) extension_name->data);
      }
    }

    if (num_enabled != (uint32_t) g_slist_length (required_extensions))
      return FALSE;

    *out_num_enabled = num_enabled;
  }

  g_free (extension_props);

  return TRUE;
}

gboolean
gulkan_instance_create (GulkanInstance *self,
                        GSList         *required_extensions)
{
  uint32_t num_enabled_layers = 0;
  char** enabled_layers = NULL;
  char** enabled_extensions =
    g_malloc(sizeof(char*) * g_slist_length (required_extensions));

  uint32_t num_enabled_extensions = 0;

  if (!_init_instance_extensions (required_extensions,
                                  enabled_extensions,
                                 &num_enabled_extensions))
    return FALSE;

  if (num_enabled_extensions > 0)
    {
      g_debug ("Requesting instance extensions:");
      for (uint32_t i = 0; i < num_enabled_extensions; i++)
        g_debug ("%s", enabled_extensions[i]);
    }

  VkResult res =
    vkCreateInstance (&(VkInstanceCreateInfo) {
      .sType = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO,
      .pApplicationInfo = &(VkApplicationInfo) {
        .sType = VK_STRUCTURE_TYPE_APPLICATION_INFO,
        .pApplicationName = "gulkan",
        .applicationVersion = 1,
        .engineVersion = 1,
        .apiVersion = VK_MAKE_VERSION (1, 0, 0)
      },
      .enabledExtensionCount = num_enabled_extensions,
      .ppEnabledExtensionNames = (const char* const*) enabled_extensions,
      .enabledLayerCount = num_enabled_layers,
      .ppEnabledLayerNames = (const char* const*) enabled_layers
    },
    NULL,
   &self->instance);

  vk_check_error ("vkCreateInstance", res, FALSE);

  for (uint32_t i = 0; i < num_enabled_extensions; i++)
    g_free (enabled_extensions[i]);
  g_free (enabled_extensions);

  return TRUE;
}

VkInstance
gulkan_instance_get_handle (GulkanInstance *self)
{
  return self->instance;
}
